/* $Id$
 *
 * Based on ttysnoop by carl (at) miskatonic.inbe.net
 */

#include <sys/select.h>
#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#define _XOPEN_SOURCE
#include <stdlib.h>

extern int errno;

static pid_t load_pid(void)
{
	FILE *fp;
	pid_t pid;

	if (!(fp = fopen("/var/run/hpcbeep.pid", "r"))) {
		fprintf(stderr, "Cannot open pidfile: %s\n"
			"Make sure that hpcbeep is running.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	if (fscanf(fp, "%u", &pid) != 1) {
		fclose(fp);
		fprintf(stderr, "Pidfile corrupt. Re-run hpcbeep.\n");
		exit(EXIT_FAILURE);
	}

	fclose(fp);

	if (kill(pid, 0) == -1) {
		fprintf(stderr, "Cannot kill pid %u: %s\n"
			"Make sure that hpcbeep is running and that you have enough privileges.\n", 
			pid, strerror(errno));
		exit(EXIT_FAILURE);
	}

	return pid;
}

static void stty_raw(int fd)
{
	struct termios term;

	tcgetattr(fd, &term);

	term.c_lflag &= ~(ICANON | IEXTEN | ISIG | ECHO);
	term.c_iflag &= ~(ICRNL | INPCK | ISTRIP | IXON | BRKINT);
	term.c_oflag &= ~OPOST;
	term.c_cflag |= CS8;

	term.c_cc[VMIN]  = 1;
	term.c_cc[VTIME] = 0;

	tcsetattr(fd, 0, &term);
}

static int open_master(void)
{
	int fd;

	fd = open("/dev/ptmx", O_RDWR);
	if (fd == -1) {
		fprintf(stderr, "Cannot open master pseudo-terminal: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	return fd;
}

static int open_slave(int master)
{
	char *name = ptsname(master);
	int fd;

	grantpt(master);
	unlockpt(master);

	fd = open(name, O_RDWR);
	if (fd == -1) {
		fprintf(stderr, "Cannot open slave pseudo-terminal: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	return fd;
}

static pid_t fork_child(int *ptyfd, char * const av[])
{
	struct termios term;
	struct winsize twin;
	int fd;
	pid_t pid;

	tcgetattr(0, &term);
	ioctl(0, TIOCGWINSZ, (char *) &twin);

	*ptyfd = open_master();

	pid = fork();
	if (pid == -1) {
		fprintf(stderr, "Cannot fork: %s\n", strerror(errno));
		close(*ptyfd);
		exit(EXIT_FAILURE);
	} else if (pid)
		return pid;

	/* child */

	fd = open_slave(*ptyfd);
	close(*ptyfd);

	setsid();
	tcsetattr(fd, TCSANOW, &term);
	ioctl(fd, TIOCSWINSZ, &twin);
	ioctl(fd, TIOCSCTTY, NULL);

	dup2(fd, 0);
	dup2(fd, 1);
	dup2(fd, 2);

	if (fd > 2)
		close(fd);

	execv(*av, av);
	exit(EXIT_FAILURE);
}

static int redirect(int srcfd, int dstfd, pid_t hbpid)
{
	char ibuf[4096], obuf[4096], *obptr = obuf;
	ssize_t rs;
	size_t ibpos, obsz;
	int beep = 0;

	rs = read(srcfd, ibuf, sizeof(ibuf));
	if (rs == -1) {
		if (errno == EAGAIN || errno == EINTR)
			return 1;

		fprintf(stderr, "read() failed: %s\n", strerror(errno));
		return 0;
	} else if (!rs)
		return 0;

	obsz = 0;
	for (ibpos = 0; ibpos < rs; ibpos++)
		if (ibuf[ibpos] == '\x07')
			beep = 1;
		else
			obuf[obsz++] = ibuf[ibpos];

	if (beep)
		kill(hbpid, SIGUSR1);

	while (obsz) {
		ssize_t wrs = write(dstfd, obptr, obsz);
		if (wrs == -1) {
			if (errno == EAGAIN || errno == EINTR)
				continue;

			fprintf(stderr, "write() failed: %s\n", strerror(errno));
			return 0;
		} else if (!wrs)
			return 0;

		obptr += wrs;
		obsz -= wrs;
	}

	return 1;
}

int main(int ac, char * const av[])
{
	pid_t hbpid, chpid;	/* hpcbeep and child pid */
	int fd;

	if (ac < 2) {
		fprintf(stderr, "Syntax: snoop program [args...]\n");
		fprintf(stderr, "Example: snoop /bin/sh\n");
		exit(EXIT_FAILURE);
	}

	if (!isatty(0)) {
		fprintf(stderr, "Must be run on a terminal\n");
		exit(EXIT_FAILURE);
	}

	signal(SIGCHLD, SIG_DFL);

	hbpid = load_pid();
	chpid = fork_child(&fd, av + 1);
	stty_raw(fd);

	for (;;) {
		int rs;
		fd_set fds;

		FD_ZERO(&fds);
		FD_SET(0, &fds);
		FD_SET(fd, &fds);

		/* xxx - should also check for writability? */
		rs = select(fd + 1, &fds, NULL, NULL, NULL);
		if (rs == -1) {
			if (errno == EAGAIN || errno == EINTR)
				continue;

			fprintf(stderr, "select() failed: %s\n", strerror(errno));
			kill(chpid, SIGKILL);	/* xxx - shouldn't be SIGHUP? */
			exit(EXIT_FAILURE);
		}

		if (FD_ISSET(0, &fds) && !redirect(0, fd, hbpid))
			break;

		if (FD_ISSET(fd, &fds) && !redirect(fd, 1, hbpid))
			break;
	}

	kill(chpid, SIGKILL);	/* sanity, should never succeed */

	return 0;
}
