/* $Id$
 *
 * hpcbeep Copyright (C) 2007 Adam Wysocki
 *
 * Send SIGUSR1 to beep with Jornada 6x0 speaker.
 *
 * Changelog:
 * - 2007-05-08
 *   - First version (gophi)
 *
 * Todo:
 * - Pidfile
 */

#include <inttypes.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <util.h>
#include <pwd.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define BEEP_SLEEP	25000
#define BEEP_COUNT	100
#define BEEP_VOLUME	0xFF

/* /usr/src/sys/arch/sh3/include/dacreg.h */

#define SH7709_DADR1		0xa40000a2
#define SH7709_DACR		0xa40000a4
#define SH7709_DACR_DAOE1	0x80
#define SH7709_DACR_DAE		0x20

/* /usr/src/sys/arch/hpcsh/dev/hd64461/hd64461gpioreg.h */

#define HD64461_GPACR_REG16	0xb0004000
#define HD64461_GPADR_REG16	0xb0004010

#define xassert(e) if (!(e)) { \
	fprintf(stderr, __FILE__ ":%d: Assertion failed: %s\n" \
		"Error: %s\n", __LINE__, #e, strerror(errno)); \
	exit(EXIT_FAILURE); \
}

volatile int flag_term = 0, flag_beep = 0;
extern int errno;

static struct
{
	int fd;
	volatile uint16_t *gpiocr;
	volatile uint16_t *gpiodr;
	volatile uint8_t *daccr;
	volatile uint8_t *dacdr;

} hpcmem;

static void privs_init(void)
{
	fprintf(stderr, "Warning: Privilege separation is still to be done.\n");
}

static void privs_raise(void)
{
}

static void privs_lower(void)
{
}

static void hpcmem_init(void)
{
	privs_raise();

	if ((hpcmem.fd = open("/dev/mem", O_RDWR)) < 0) {
		privs_lower();
		fprintf(stderr, "Cannot open /dev/mem: %s\n"
			"Make sure that you have securelevel turned to 0.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	privs_lower();

	xassert((hpcmem.gpiocr = mmap(NULL, sizeof(*hpcmem.gpiocr), 
		PROT_READ | PROT_WRITE, MAP_SHARED, hpcmem.fd, HD64461_GPACR_REG16)) != MAP_FAILED);

	xassert((hpcmem.gpiodr = mmap(NULL, sizeof(*hpcmem.gpiodr), 
		PROT_READ | PROT_WRITE, MAP_SHARED, hpcmem.fd, HD64461_GPADR_REG16)) != MAP_FAILED);

	xassert((hpcmem.daccr = mmap(NULL, sizeof(*hpcmem.daccr), 
		PROT_READ | PROT_WRITE, MAP_SHARED, hpcmem.fd, SH7709_DACR)) != MAP_FAILED);

	xassert((hpcmem.dacdr = mmap(NULL, sizeof(*hpcmem.dacdr), 
		PROT_READ | PROT_WRITE, MAP_SHARED, hpcmem.fd, SH7709_DADR1)) != MAP_FAILED);

	*hpcmem.daccr = ((*hpcmem.daccr) & (~SH7709_DACR_DAE)) | SH7709_DACR_DAOE1;
}

static void hpcmem_cleanup(void)
{
	*hpcmem.daccr &= ~(SH7709_DACR_DAE | SH7709_DACR_DAOE1);

	munmap((void *) hpcmem.dacdr, sizeof(*hpcmem.dacdr));
	munmap((void *) hpcmem.daccr, sizeof(*hpcmem.daccr));
	munmap((void *) hpcmem.gpiodr, sizeof(*hpcmem.gpiodr));
	munmap((void *) hpcmem.gpiocr, sizeof(*hpcmem.gpiocr));

	close(hpcmem.fd);
}

static void spk_on(void)
{
	*hpcmem.gpiodr &= ~0x01;
}

static void spk_off(void)
{
	*hpcmem.gpiodr |= 0x01;
}

static void spk_send(uint8_t sample)
{
	*hpcmem.dacdr = sample;
}

static void sighnd(int signo)
{
	if (signo == SIGUSR1)
		flag_beep = 1;
	else if (signo == SIGTERM || signo == SIGINT)
		flag_term = 1;

	signal(signo, sighnd);
}

static void beep(void)
{
	size_t i;
	uint8_t v = 0;

	spk_on();

	for (i = 0; i < BEEP_COUNT; i++) {
		size_t sl;

		spk_send(v);
		v ^= BEEP_VOLUME;

		/* System calls calling is soo slow that we must busy-wait. I don't 
		 * like this idea too :( */
		for (sl = 0; sl < BEEP_SLEEP; sl++)
			__asm__(("nop"));
	}

	spk_send(0x00);
	spk_off();
}

int main(void)
{
	privs_init();

	hpcmem_init();
	spk_send(0x00);
	spk_off();

	signal(SIGTERM, sighnd);
	signal(SIGINT, sighnd);
	signal(SIGUSR1, sighnd);

	privs_raise();
	xassert(setpriority(PRIO_PROCESS, 0, -19) != -1);
	privs_lower();

	errno = 0;
	if (daemon(0, 0) == -1) {
		fprintf(stderr, "Cannot daemonize: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	if (pidfile(NULL) == -1) {
		fprintf(stderr, "Cannot write pidfile: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	while (!flag_term) {
		pause();

		if (flag_beep) {
			flag_beep = 0;
			beep();
		}
	}

	spk_off();
	hpcmem_cleanup();

	signal(SIGUSR1, SIG_DFL);
	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);

	return 0;
}
